
/*
   helifile.c
   
   File I/O routines for Helicon.

   Author: Mark E. Nash
   Date:   June, 1988
*/

#include	<sys\types.h>
#include	<sys\timeb.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<malloc.h>
#include	<string.h>
#include	<process.h>
#include	<graph.h>
#include	<conio.h>
#include	<io.h>
#include	<dos.h>
#include	<bios.h>
#include	"helimacs.h"
#include	"helimess.h"
#include	"helstruc.h"
#include	"helfdef.h"


extern char msg[];				/* in heliglob.h */
extern struct grafwindow *top_window;		/* in heliglob.h */
extern unsigned delcon;				/* in heliglob.h */
extern short cxpos, cypos;			/* in heliglob.h */
extern unsigned mouse;				/* in heliglob.h */
extern unsigned m1, m2, m3, m4;			/* in heliglob.h */
extern struct mimage mbar;			/* in heliglob.h */
extern int prtyp;				/* in heliglob.h */
extern char tboxes[NUMTBXS][TBOXSIZE+1];	/* in heliglob.h */
extern char clrstr[];				/* in heliglob.h */
extern int lastnum;				/* in heliglob.h */
extern double params[MAXPARS+1][NUMDIST+1];	/* heliglob.h */



int hel_init()
/*
   hel_init() does initialialization for helicon video, graphics, and other
   things.

   Entry: Helicon needs to be initialized.
   
   Exit:  Returns NOERR if ok, exits if error.
*/
	{
	extern struct videoconfig far vconfig;		/* for video config */
	struct timeb stimeb, ftimeb;			/* for timer const */
	unsigned long int tloop;			/* for timer const */
	extern unsigned long mfree;			/* in heliglob.h */
	extern unsigned long mopap;			/* in heliglob.h */
	extern unsigned long mtot;			/* in heliglob.h */
	extern unsigned rseed;				/* in heliglob.h */
	unsigned msize;				/* for getting memory size */
	int	ctr;				/* gp counter */
	

/* get the total memory size */
	mtot = ((unsigned long)_bios_memsize()) * 1024l;

/* find out the free memory size by allocating truck loads of memory and
   see how much we get */
   	_dos_allocmem(MAXINT, &msize);
	_dos_freemem(msize);
	mfree = 16l * (unsigned long)msize;
	
/* find out the operating system size */
	mopap = mtot - mfree;
	
/* initialize some globals */
	strcpy(tboxes[DATFIL], "");
	sprintf(msg, "%d", NUMBINS);
	strcpy(tboxes[NBINS], msg);
	strcpy(tboxes[BVLU], "1.0");	/* initial b-value in weibull */
	strcpy(tboxes[EVLU], "0.000001");	/* weicalc() iterative error */
	
/* initialize random number generator */
	strcpy(tboxes[SEED], "0");
	rseed = (unsigned)atoi(tboxes[SEED]);
	srand(rseed);

	for (ctr=NULLDIST+1; ctr<=NUMDIST; ctr++)
		params[MSE][ctr] = -1.0;

/* set modes */
	if (_setvideomode(_ERESCOLOR) == 0)
		{
		printf(I_NOEGA);
		return(ERR);
		}
	
	_getvideoconfig(&vconfig);	/* set up video config structure */
	if (_setactivepage(0) < 0)	/* set page to write to */
		{
		printf(I_SAPN);
		return(ERR);
		}
	if (_setvisualpage(0) < 0)	/* set page to display */
		{
		printf(I_SVPN);
		return(ERR);
		}
			
/* put logical origin in center */
	_setlogorg(vconfig.numxpixels/2 - 1, vconfig.numypixels/2 - 1);
	
/* null out window list */
	top_window = NULL;

/* find the window zoom-border timer delay.  This will make the zoom box
   machine independent. */
	ftime(&stimeb);
	
#pragma loop_opt(off)
	for (tloop=0; tloop<DELMAX; tloop++);
#pragma loop_opt()

	ftime(&ftimeb);
	delcon = (unsigned)((double)ZSPEED / (double)(ftimeb.millitm + 1000 * 
				(ftimeb.time - stimeb.time) - stimeb.millitm));

/* check for mouse */
	mouse = TRUE;		/* turn it on for the check */
	MOUSE(0, 0, 0, 0);
	if (m1 != -1)
		mouse = FALSE;

	return(NOERR);
	}
	
	

int hel_exit()
/*
   hel_exit() exits from helicon.

   Entry: Helicon needs to be exited.
   
   Exit:  Back to dos.
*/
	{
/* finalize things */
	egareset();			/* reset the ega registers */
	_setvideomode(_DEFAULTMODE);	/* set the video mode */
	exit(NOERR);
	}
	
	
	
int readatin()
/*
   readatin() reads in the data from the raw data file.
   
   Entry: A read has been requested.
          The global datfil points to the file name.
	   
   Exit:  If no errors, the file has been read into the rawdat list.
*/
	{
	extern char tboxes[NUMTBXS][TBOXSIZE+1];	/* in heliglob.h */
	extern struct rawcell *rd_top;			/* in heliglob.h */
	struct grafwindow *ms;				/* for messages */
	FILE	*fp;					/* files */


/* try to access input file */
	if (access(tboxes[DATFIL], 00) == -1)
		{
		sprintf(msg, F_FNF, tboxes[DATFIL]); 
		err_window(msg);
		return(NOERR);
		}

/* try to open the input file */
	if ((fp = fopen(tboxes[DATFIL], "r")) == NULL)
		{
		sprintf(msg, F_COPEN, tboxes[DATFIL]);
		err_window(msg);
		return(NOERR);
		}

/* display read-in message */
	sprintf(msg, I_RDFILE, tboxes[DATFIL]);
	ms = msg_window(msg);

	free_rawdat(&rd_top);
	read_rawdat(fp, &rd_top);
	lastnum = 0;		/* reset so that current pars are invalid */
	
/* get rid of message and finish up */
	close_grafwin(&ms);
	fclose(fp);
	return(NOERR);
	}
	
	
	
int read_rawdat(fp, top)
	FILE *fp;
	struct rawcell **top;
/*
   read_rawdat() reads and sorts the raw data file.
   
   Entry: fp is the file to read from, and it exists.
	  
   Exit:  *top points to the top of the raw data list.
          If no errors, the list has been read in,
          else returns ERR.
*/
	{
	extern char msg[];		/* in heliglob.c */
	char	string[BIGSTRING];	/* for file input */
	char	tmp;			/* for null insertion */
	int	ctr;			/* line counter for input */
	char	*st1, *st2;		/* string pointers used for parsing */
	struct rawcell *rd, *prev;	/* sorting pointers */
	struct rawcell *rdtop;		/* top raw data in the list */
	struct rawcell *data;		/* data that has been read */
	double	dvalue;			/* domain value */
	int	valid;			/* is the data valid? */


	*top = NULL;
/* get the first valid line of data */
	rdtop = NULL;
	ctr = 1;	/* skip first line */
	while(!feof(fp) && !(valid = gethdat(fp, &dvalue, ctr)))
		ctr++;
		
	if (valid)
		{
		if ((rdtop = (struct rawcell *)malloc(sizeof(struct rawcell))) 
								== NULL)
			{
			printf("\7");
			return(ERR); 
			}
/* update memory status */
		memupdt((unsigned long)sizeof(struct rawcell), ALLOC);

		rd = rdtop;
		rd->x = dvalue;
		rd->next = NULL;
		}
	else
		{
		err_window(F_NODAT);
		free_rawdat(&rdtop);
		return(ERR);
		}

/* get the following lines of data and sort */
	rd = rdtop;
	prev = NULL;
	while (!feof(fp))
		{
		ctr++;
		valid = gethdat(fp, &dvalue, ctr);
		if (valid)
			{
	/* allocate storage for the data read */
			if ((data = (struct rawcell *)malloc
					(sizeof(struct rawcell))) == NULL)
				{
				printf("\7");
				return(ERR);
				}
/* update memory status */
			memupdt((unsigned long)sizeof(struct rawcell), ALLOC);

			data->x = dvalue;
			data->next = NULL;
			
			if (dvalue < rd->x)	/* start at beginning */
				{
				rd = rdtop;
				prev = NULL;
				if (dvalue < rdtop->x)	/* new start */
					{
					data->next = rdtop;
					rdtop = data;
					}

				else			/* search for start */
					{
					while (rd->next != NULL && 
							dvalue >= rd->x)
						{
						prev = rd;
						rd = rd->next;
						}
					data->next = rd;
					prev->next = data;
					}
				}

			else		/* insert at interior or end */
				{
		/* get to the spot we want */
				while (rd->next != NULL && dvalue >= rd->x)
					{
					prev = rd;
					rd = rd->next;
					}
					
				if (rd->x > dvalue) /* can happen at end of list */
					{
					prev->next = data;
					data->next = rd;
					prev = rd;
					}
				else
					{
					rd->next = data;
					prev = rd;
					rd = rd->next;
					}
				}
			}
		}

/* assign the list */
	*top = rdtop;
			
	return(NOERR);
	}
	
	
	
int gethdat(fp, dvalue, ctr)
	FILE *fp;
	double *dvalue;
	int ctr;
/*
   gethdat() gets a line of helicon format data from the input file.
   
   Entry: fp is the file to read from.
          ctr is the line number that is currently being read.
   
   Exit:  Returns TRUE if the data is valid.
          *dvalue is the domain value, if valid.
*/
	{
	extern char msg[];		/* in heliglob.c */
	char	string[BIGSTRING];	/* for file input */
	char	tmp;			/* for null insertion */
	char	*st1, *st2;		/* string pointers used for parsing */
	int	dvalid, valid;		/* tells whether data read is valid */
	
	
	dvalid = FALSE;
	valid = TRUE;
	if (fgets(string, BIGSTRING, fp) == NULL)
		valid = FALSE;
	else
		st1 = string;

/* get the domain variable */
	if ((st2 = pstring(&st1)) == NULL)
		{
		sprintf(msg, F_DDAT, ctr);
		err_window(msg);
		}
	else
		{
		tmp = *st2;
		*st2 = '\0';		/* temporary null insertion */
		*dvalue = atof(st1);
		*st2 = tmp;
		dvalid = TRUE;
		}

	if (valid && dvalid)
		return(TRUE);
	else
		return(FALSE);
	}
	
	
	
char *pstring(beg)
	char **beg;
/*
   pstring() parses a variable name out of a string.
   
   Entry: *beg points to the parsing starting point of the string.
	  
   Exit:  *beg points to the beginning of the parsed string (if one was found).
          If a string was found, returns a pointer to one character past the
	  	end of the string,
          else returns NULL if no string was found.
	  
   Note:  The parsing rule is to look for spaces.  It will skip over leading
          white space to the first character.  It will then look for the next
	  white space or null character to use as the end of the string.
*/
	{
	char *end;		/* end of the parsed string */
	long int len;		/* length of white space */
	long int ctr;		/* gp counter */
	
	
/* skip over any leading white space */
	len = (long)strspn(*beg, WHITESPACE);
	for (ctr=0; ctr<len; ctr++)
		(*beg)++;

/* get the valid characters */
	end = *beg;
	end = strpbrk(*beg, WHITESPACE);
	
	return(end);
	}
	
	
	
int free_rawdat(top)
	struct rawcell **top;
/*
   free_rawdat() liberates the raw data list.
   
   Entry: *top points to the top of the list.
	  
   Exit:  *top points to NULL.
          If no errors, the list has been distroyed,
          else returns ERR.
*/
	{
	struct rawcell *rd, *temp;
	
	
	if (*top == NULL)
		return(NOERR);

	rd = *top;
	while (rd != NULL)
		{
		temp = rd;
		rd = rd->next;
		free((void *)temp);
/* update memory status */
		memupdt((unsigned long)sizeof(struct rawcell), FRE);
		}
		
	*top = NULL;

/* check the heap */
		switch (_heapchk())
			{
			case _HEAPOK:
			case _HEAPEMPTY:
				break;
			case _HEAPBADBEGIN:
			case _HEAPBADNODE:
				printf("\7");
				return(ERR);
			}
	return(NOERR);
	}
	
	
	
int err_window(errmsg)
	char errmsg[];
/*
   err_window() displays an 80-character two-line error message and a prompt
   line in a text window.
   
   Entry: errmsg[] contains the error message to display.
   
   Exit:  If no errors, the message has been displayed, and the user 
   		has responded.
*/
	{
	struct grafwindow *ew;			/* error window */
	register unsigned int ctr;		/* jeep counter */
	struct xycoord xy;			/* coord conv */
	int	done;				/* mouse loop */
	
	
	if ((ew = open_grafwin(TEXTWIN, _getlogcoord(ERRWX1, ERRWY1).xcoord,
		                       _getlogcoord(ERRWX1, ERRWY1).ycoord,
		   	0, 0, ERRBORDER, ERRBACK, ERRCOLS, ERROWS, TRUE))
								== NULL)
		return(ERR);
	
	eputs_window(ew, errmsg, EMESCOLOR, TRUE);
	
/* Display prompt message */
	gtwcurpos(ew, EPCOL, EPROW);
	if (mouse)
		eputs_window(ew, MERRPROMP, EPRCOLOR, TRUE);
	else
		eputs_window(ew, ERRPROMP, EPRCOLOR, TRUE);
	
	if (mouse)
		{
/* check the mouse status */
		done = FALSE;
		while (!done)
			{
			MOUSE(3, 0, 0, 0);
			xy = _getlogcoord(m3, m4);
			cxpos = xy.xcoord;
			cypos = xy.ycoord;
			if (m2 != 0)
				if (curnrgn(&(ew->wbox)))
					done = TRUE;
			}
		}
/* wait for a return character */
	else
		while (getch() != RETURN);
	
	close_grafwin(&ew);
	return(NOERR);
	}
	
	
	
struct grafwindow *msg_window(message)
	char message[];
/*
   msg_window() displays an 80-character two-line informational message.
   
   Entry: message[] contains the error message to display.
   
   Exit:  If no errors, the message has been displayed.
          Returns a pointer to the message window for closing.
*/
	{
	struct grafwindow *ew;			/* error window */
	
	
	if ((ew = open_grafwin(TEXTWIN, _getlogcoord(ERRWX1, ERRWY1).xcoord,
		                       _getlogcoord(ERRWX1, ERRWY1).ycoord,
			   0, 0, MSGBORDER, MSGBACK, MSGCOLS, MSGROWS, TRUE))
								== NULL)
		return(ERR);
	
	eputs_window(ew, message, MESCOLOR, TRUE);
	
	return(ew);
	}
	
	
	
int egaprnt()
/*
   egaprnt() prints out the helicon screen.
   
   Entry: A print request has been made.
   
   Exit:  The entire screen has been printed out (except the menus).
*/
	{
	struct grafwindow *ps;		/* print status window */
	struct grafwindow *pg;		/* print status graphics window */
	struct grafwindow *pb;		/* print status button window */
	struct xycoord xy1, xy2;	/* coord conv */
	char far *img;			/* back of print window pointer */
	long	siz;			/* size of the background image */
	
	
/* initialize the printer */
	if (prinit() == ERR)
		return(ERR);

	xy1 = _getlogcoord(SWX1, SWY1);

/* get our copy of the print stat window background */
	geti(xy1.xcoord, xy1.ycoord, xy1.xcoord+HPIXCHAR*(SWCOLS+1),
				xy1.ycoord+VPIXCHAR*(SWROWS+1), &img, &siz);

/* open the printer status window */
	if ((ps = open_grafwin(TEXTWIN, xy1.xcoord, xy1.ycoord, 0, 0, 
			STBRCOLOR, STBKCOLOR, SWCOLS, SWROWS, TRUE)) == NULL)
		return(ERR);

/* open the graphics window */
	xy1 = _getlogcoord(SGWX1, SGWY1);
	xy2 = _getlogcoord(SGWX2, SGWY2);
	if ((pg = open_grafwin(GRAFWIN, xy1.xcoord, xy1.ycoord, xy2.xcoord, 
			xy2.ycoord, STGBRCOLOR, STGBKCOLOR, 0, 0, FALSE)) 
								== NULL)
		return(ERR);

/* open the button window */
	xy1 = _getlogcoord(SBWX1, SBWY1);
	if ((pb = open_grafwin(BWIN, xy1.xcoord, xy1.ycoord, 0, 0, SBBRCOLOR, 
				SBBKCOLOR, SBWCOLS, SBWROWS, FALSE)) == NULL)
		return(ERR);

/* display window text */
	eputs_window(ps, "Print Status       Completed   0%", PSSCOLOR, FALSE);
	eputs_window(pb, "Cancel", PBSCOLOR, FALSE);

	scale_grafwin(pg, 0.0, 0.0, 100.0, 3.0);
	win_dscale(pg, PGSCACOLOR, XAXIS);
	
/* put the mouse cursor in the window */
	MOUSE(4, 0, SWX1+HPIXCHAR*(SWCOLS-1), SWY1+VPIXCHAR*(SWROWS-3));

/* print the screen */
	switch (prtyp)
		{
		case HPLASJET:
			hpscrprn(ps, pb, pg, SWX1, SWY1, SWX1+HPIXCHAR*
				    (SWCOLS+1), SWY1+VPIXCHAR*(SWROWS+1), img);
			break;
		case EPSON:
			epscrprn(ps, pb, pg, SWX1, SWY1, SWX1+HPIXCHAR*
				    (SWCOLS+1), SWY1+VPIXCHAR*(SWROWS+1), img);
			break;
		}
		
/* free up the image memory */
	ifree(&img, siz);
	
/* close the windows */
	close_grafwin(&pb);
	close_grafwin(&pg);
	close_grafwin(&ps);
	
	return(NOERR);
	}
	
	
	
int prinit()
/*
   prinit() initializes the printer.
   
   Entry: The printer needs to be initialized.
   
   Exit:  If the printer is supported, it has been initialized,
   		and the global prtyp has been set,
   	  else, returns ERR.
*/
	{
	char	prntr[10];			/* printer command string */
	
	
/* get the printer type if it is not already set */
	switch (prtyp)
		{
		case HPLASJET:
			prntr[0] = 27;		/* resets the printer */
			prntr[1] = 'E';
			prntr[2] = 0;
			fprintf(stdprn, "%s", prntr);

			prntr[1] = '&';		/* selects landscape mode */
			prntr[2] = 'l';
			prntr[3] = '1';
			prntr[4] = 'O';
			prntr[5] = 0;
			fprintf(stdprn, "%s", prntr);

			prntr[1] = '&';		/* set top margin */
			prntr[2] = 'l';
			prntr[3] = '1';
			prntr[4] = 'E';
			prntr[5] = 0;
			fprintf(stdprn, "%s", prntr);

			prntr[1] = '*';	/* raster graphics resolution */
			prntr[2] = 't';
			prntr[3] = '7';		/* 75 dpi */
			prntr[4] = '5';
			prntr[5] = 'R';
			prntr[6] = 0;
			fprintf(stdprn, "%s", prntr);

					/* move to column centered */
		/* length of page is 10.6 inches, 720 decipoints per inch */
			prntr[1] = '&';
			prntr[2] = 'a';
			prntr[3] = '0' + 6;
			prntr[4] = '0' + 8;
			prntr[5] = '0' + 8;
			prntr[6] = '0' + 8;
			prntr[7] = 'H';
			prntr[8] = 0;
			fprintf(stdprn, "%s", prntr);

					/* move to row centered */
		/* width of page is 8.0 inches, 720 decipoints per inch */
			prntr[1] = '&';
			prntr[2] = 'a';
			prntr[3] = '0' + 1;
			prntr[4] = '0' + 2;
			prntr[5] = '0';
			prntr[6] = '0';
			prntr[7] = 'V';
			prntr[8] = 0;
			fprintf(stdprn, "%s", prntr);

			prntr[1] = '*';		/* start raster graphics */
			prntr[2] = 'r';
			prntr[3] = '1';
			prntr[4] = 'A';
			prntr[5] = 0;
			fprintf(stdprn, "%s", prntr);
			break;
			
		case EPSON:
			prntr[0] = 27;
			prntr[1] = '3';
			prntr[2] = 24;
			prntr[3] = 0;
			fprintf(stdprn, "%s", prntr);
			break;
			
		case NULLPT:
			sprintf(msg, R_NOSUP, tboxes[PRTYPE]);
			err_window(msg);
			return(ERR);
			
		default:
			sprintf(msg, R_BADTYP, prtyp, "prinit()");
			err_window(msg);
			return(ERR);
		}
	return(NOERR);
	}
	
	
	
int printyp()
/*
   printyp() gets a numerical value for the printer type.
   
   Entry: The proper tboxes entry contains the printer type as a string.
   
   Exit:  A printer type has been returned.
*/
	{
	int	pt;			/* printer type */
	char	prt[TBOXSIZE+1];	/* printer type as a string */
	
	
	pt = NULLPT;
	strncpy(prt, tboxes[PRTYPE], TBOXSIZE);
	strlwr(prt);

/* compare to all supported printers */
	if (strcmp(prt, HPSUP) == 0)
		pt = HPLASJET;
	if (strcmp(prt, EPSUP) == 0)
		pt = EPSON;

	if (pt == NULLPT)
		{
		sprintf(msg, R_NOSUP, tboxes[PRTYPE]);
		err_window(msg);
		}
		
	return(pt);
	}
	
	
	
int hpscrprn(tw, bw, gw, x1, y1, x2, y2, back)
	struct grafwindow *tw, *bw, *gw;
	short x1, y1, x2, y2;
	char far *back;
/*
   hpscrnprn() prints the screen on an hp laser jet printer.
   
   Entry: tw points to the printer text window,
          bw points to the printer button window,
	  gw points to the printer graphics window,
          (x1,y1), (x2,y2) are the print window bounds,
          back points to the print window background image.
	  
   Exit:  The screen has been printed.
*/
	{
	int	row, col;		/* row and column pixel counters */
	char	pixbyte[400 / 8 + 2];	/* holds one printable row */
	char	dotvalue;		/* one pixel's value */
	int	bytecnt;		/* byte counter */
	int	pixnum;			/* pixel counter */
	char	initstr[7];		/* printer string */
	int	rows, cols;		/* constants for counters */
	int	bi;		/* byte index for sending string to printer */
	double	pctdone, pctchk;		/* percent done for status */
	int	coldone;		/* number of columns done */
	double	xbeg, xend;		/* for updating print bar */
	int	key;			/* for cancel button */
	struct xycoord xy;		/* for coord conv */


/* set up loop limits */
	rows = EGAPIXROWS - 1;
	cols = EGAPIXCOLS - 1;


	initstr[0] = 27;       /* control code to set # of bytes raster data */
	initstr[1] = '*';
	initstr[2] = 'b';

				/* how many bytes per row to print */
	initstr[3] = (char)('0' + (char)(((rows / 8) + 1) / 10));
	initstr[4] = (char)('0' + (char)(((rows / 8) + 1) % 10));

	initstr[5] = 'W';
	initstr[6] = 0;

	coldone = 0;
	pctdone = 0.0;
	for (col = cols; col >= 0; col--)	/* for each column . . . */
		{
		bytecnt = 0;
		pixbyte[bytecnt] = '\0';
		pixnum = 0;

	/* check for cancel button */
		/* check key */
		if (kbhit())
			{
			key = getch();
			if (key == ESC)
				{
				hilbtn(bw);
				laserend();
				return(NOERR);
				}
			}

		for (row = 0; row <= rows; row++)	/* process each row */
			{
	/* check for cancel button */
		/* check mouse */
			MOUSE(3, 0, 0, 0);
			xy = _getlogcoord(m3, m4);
			cxpos = xy.xcoord;
			cypos = xy.ycoord;
			if (curnrgn(&(bw->wbox)))
				if (m2 != 0)
					{
					hilbtn(bw);
					laserend();
					return(NOERR);
					}

	/* read in the pixel's value */
			if (col >= x1 && col <= x2 && row >= y1 && row <= y2)
				{
				dotvalue = (char)ctob(*back);
				back++;
				}
			else
				dotvalue = (char)ctob((char)getpt(col, row));


			pixbyte[bytecnt] <<= 1;
			pixbyte[bytecnt] |= dotvalue;
			pixnum++;	/* count 8 bits per byte */

			if (pixnum == 8)	/* byte is complete */
				{
				bytecnt++;	/* increment byte count */
				pixbyte[bytecnt] = '\0';
				pixnum = 0;	/* restart pixel count */
				}
			}

					/* slide last byte into position */
		if (pixnum != 0)
			pixbyte[bytecnt] <<= (8 - pixnum);

				/* send that row to the printer */
		fprintf(stdprn, "%s", initstr);
		for (bi = 0; bi < (rows / 8) + 1; bi++)
			fprintf(stdprn, "%c", pixbyte[bi]);

/* update status window */
		coldone++;
		pctchk = 100.0 * (double)coldone / (double)EGAPIXCOLS;
		if ((int)(pctchk+0.5) != (int)(pctdone+0.5))
			{
/* update text */
			sprintf(msg, "%3d", (int)(pctchk+0.5));
			gtwcurpos(tw, PSWUC, PSWUR);
			clrstr[3] = '\0';
			eputs_window(tw, clrstr, STBKCOLOR, FALSE);
			clrstr[3] = '\200';
			gtwcurpos(tw, PSWUC, PSWUR);
			eputs_window(tw, msg, PSSCOLOR, FALSE);
			}

/* update graphics */
		win_rect(gw, pctdone, PRIY1, pctchk, PRIY2, STGDCOLOR, TRUE);
		pctdone = pctchk;
		}

	laserend();
	
	return(NOERR);
	}
	
	
	
int epscrprn(tw, bw, gw, x1, y1, x2, y2, back)
	struct grafwindow *tw, *bw, *gw;
	short x1, y1, x2, y2;
	char far *back;
/*
   epscrnprn() prints the screen on an epson-type printer.
   
   Entry: tw points to the printer text window,
          bw points to the printer button window,
	  gw points to the printer graphics window,
          (x1,y1), (x2,y2) are the print window bounds,
          back points to the print window background image.
	  
   Exit:  The screen has been printed.
*/
	{
	int	row, col;		/* row and column pixel counters */
	char	pixbyte[401];		/* holds one printable row */
	char	dotvalue;		/* one pixel's value */
	int	colcnt, ccol;		/* column counters */
	int	pixnum;			/* pixel counter */
	char	initstr[7];		/* printer string */
	char	crlf[3];		/* carriage return-line feed string */
	int	rows, cols;		/* constants for counters */
	int	bi;		/* byte index for sending string to printer */
	double	pctdone, pctchk;		/* percent done for status */
	int	coldone;		/* number of columns done */
	double	xbeg, xend;		/* for updating print bar */
	int	key;			/* for cancel button */
	struct xycoord xy;		/* for coord conv */
	long	rowptr, colptr, browsize, bcolsize;/* to convert back info */
	char far *curpix;		/* current pixel to get */


					/* set up loop limits */
	rows = EGAPIXROWS - 1;
	cols = EGAPIXCOLS - 1;

	crlf[0] = 10;		/* line feed */
	crlf[1] = 13;		/* carriage return */
	crlf[2] = 0;

	initstr[0] = 27;	/* string to tell the epson fx-80 to */
	initstr[1] = 'K';	/* print in single density graphics mode */

				/* how many bytes per row to print */
	initstr[2] = (char)((rows + 1) % 256);
	initstr[3] = (char)((rows + 1) / 256);
	initstr[4] = 0;

				/* for each set of 8 columns */
	coldone = 0;
	pctdone = 0.0;
	browsize = (long)(y2-y1+1);
	bcolsize = (long)(x2-x1+1);
	for (colcnt = 0; colcnt <= cols; colcnt += 8)
		{
	/* check for cancel button */
		/* check key */
		if (kbhit())
			{
			key = getch();
			if (key == ESC)
				{
				hilbtn(bw);
				fprintf(stdprn, FORMFEED);
				fflush(stdprn);
				return(NOERR);
				}
			}
					/* for each row */
		for (row = 0; row <= rows; row++)
			{
	/* check for cancel button */
		/* check mouse */
			MOUSE(3, 0, 0, 0);
			xy = _getlogcoord(m3, m4);
			cxpos = xy.xcoord;
			cypos = xy.ycoord;
			if (curnrgn(&(bw->wbox)))
				if (m2 != 0)
					{
					hilbtn(bw);
					fprintf(stdprn, FORMFEED);
					fflush(stdprn);
					return(NOERR);
					}

			pixbyte[row] = 0;
						/* process each column */
			for (col = 0; col <= 7; col++)
				{
				ccol = col + colcnt;
				
/* read in the pixel's value */
				if (ccol >= x1 && ccol <= x2 && row >= y1 && 
								row <= y2)
					{
					curpix = (char far *)((long)back+
						(browsize*(bcolsize-1))-
						(browsize*(long)(ccol-x1))+
						(long)(row-y1));
					dotvalue = (char)ctob(*curpix);
					}
				else
					dotvalue = (char)ctob(getpt(ccol, 
									row));
							
				pixbyte[row] <<= 1;
				pixbyte[row] |= dotvalue;
				}
			}

		fprintf(stdprn, "%s", initstr); 
		for (bi = rows; bi >= 0; bi--)	/* send bytes bottom to top */
			fprintf(stdprn, "%c", pixbyte[bi]);

		fprintf(stdprn, "%s", crlf);

/* update status window */
		coldone += 8;
		pctchk = 100.0 * (double)coldone / (double)EGAPIXCOLS;
		if ((int)(pctchk+0.5) != (int)(pctdone+0.5))
			{
/* update text */
			sprintf(msg, "%3d", (int)(pctchk+0.5));
			gtwcurpos(tw, PSWUC, PSWUR);
			clrstr[3] = '\0';
			eputs_window(tw, clrstr, STBKCOLOR, FALSE);
			clrstr[3] = '\200';
			gtwcurpos(tw, PSWUC, PSWUR);
			eputs_window(tw, msg, PSSCOLOR, FALSE);
			}

/* update graphics */
		win_rect(gw, pctdone, PRIY1, pctchk, PRIY2, STGDCOLOR, TRUE);
		pctdone = pctchk;
		}

	fprintf(stdprn, FORMFEED);
	fflush(stdprn);

	return(NOERR);
	}
	
	
	
int hilbtn(hw)
	struct grafwindow *hw;
/*
   hilbtn() hilites a button that has been "pressed".
   
   Entry: hw points to a button window.
   
   Exit:  The button window has been hilited.
*/
	{
	short	xp, yp;		/* phys coord of center of button for fill */
	unsigned delctr1, delctr2;	/* for cancel button hilite delay */
	
	
	if (hw == NULL)
		return(NOERR);
		
	if (hw->wintype != BWIN)
		return(NOERR);
		
/* turn mouse off */
	MOUSE(2, 0, 0, 0);
	
/* find the center of the button */
	xp = hw->wbox.x1 + abs(hw->wbox.x2 - hw->wbox.x1) / 2;
	yp = hw->wbox.y1 + abs(hw->wbox.y2 - hw->wbox.y1) / 2;

/* hilight the button area */
	_setfillmask(NULL);
	_setcolor(BHILCOLOR);
	_floodfill(xp, yp, SBBRCOLOR);

/* delay loop */
#pragma loop_opt(off)
		for (delctr2=0; delctr2<BHIMAX; delctr2++)
			for (delctr1=0; delctr1<delcon; delctr1++);
#pragma loop_opt()
	

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);
	
	return(NOERR);
	}
	
	
	
int laserend()
/* 
   laserend() ejects the page from the HP LaserJet after dumping graphics.
*/
	{
	char prntr[6];				/* printer command string */


	prntr[0] = 27;
	prntr[1] = '*';
	prntr[2] = 'r';
	prntr[3] = 'B';
	prntr[4] = 0;
	fprintf(stdprn, "%s", prntr);

	prntr[0] = 27;
	prntr[1] = 'E';
	prntr[2] = 0;
	prntr[3] = 0;
	fprintf(stdprn, "%s", prntr);
	
/* flush the standard printer stream */
	fflush(stdprn);
	
	return(NOERR);
	}
